#!/usr/bin/env perl

################################################################################
#
# Copyright (C) 2011 Sarah Eris Horsley Caffee
#
# This is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
#
# Returns a new unused username for use on a *nix server.
#
################################################################################

use warnings;
use strict;
use English;
use File::Basename;

my $base ="";
my $user;
my $debug = 0;
my $maxlen = 8;
my $test = 0;

foreach my $arg (@ARGV) {
    if ("-d" eq $arg) {
	$debug = 1;
    } elsif ($arg =~ /^-m/) {
	$arg =~ s/^..//;
	if ($arg =~ /^(\d+)$/) {
	    $maxlen = $1;
	    $debug and print "Setting maxlen to $maxlen\n";
	} else {
	    printf(STDERR "Invalid username length specified: $arg\n");
	}
    } elsif ($arg eq "-h") {
	usage();
	exit 0;
    } elsif ($arg eq "-t") {
	$test = 1;
    } else {
	$base = $arg;
    }
}

if (! $test) {
    $user = make_username($base);
    defined $user and printf("%s\n", $user);
    exit 0;
} else {
    my $pwdfile = "passwd";
    $base = "";
    my $t1 = time;
    $user = make_username($base, $pwdfile);
    my $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);
    
    $base = "eightchr";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);
    
    $base = "short";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);
    
    $base = "this *&% is not allowed &^$%&^m";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);

    $base = "123digits";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);

    $base = "this-is-a-domain.com";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);

    $base = '&$%#&$*&$%*&';
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);

    $base = "testqwer";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);

    $base = "testabc";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);

    $base = "testzx";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);

    $base = "test";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);

    $base = "t";
    $t1 = time;
    $user = make_username($base, $pwdfile);
    $t2 = time;
    defined $user and printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);
}

sub make_username {
    # Arguments:
    # The string to use as the basis of the new username.
    # Can be any string (such as an existing username or a domain name)
    # But at most 8 characters will be used.  Invalid characters will
    # be removed.
    # If the first character would be a number, an 'a' is prepended to the name.
    # If blank, the string "user" will be used.
    #
    # Returns:
    # A string containing the new username or undef if no username 
    # could be generated  (unlikely)

    my ($base, $pwdfile) = @_;

    if ($base eq "") {
	$base = "user";
    }

    ! defined $pwdfile and $pwdfile = "/etc/passwd";

    if (! -e $pwdfile) {
	printf(STDERR "Error: non-existant passwd file specified: $pwdfile\n");
	return undef;
    }

    # Remove non-alpha-numeric characters and make sure the first character 
    # is a letter.
    $base =~ s/[^A-Za-z0-9]//g;
    if ($base !~ /^[a-z]/ ) { 
        $base = "a".$base;
    }

    # Use the specified parameter as the base of the new username.
    # Print it into a string of exactly 8 characters.
    # Print the number into the last portion of the new username.
    # Delete spaces (in case the passed username candidate was short).
    # Churn until we find an unused name.
    my $tries=0;
    my $i=1;
    my $numstr;
    my $newuser = sprintf("%.".$maxlen."s", $base);

    while ( (user_exists($newuser, $pwdfile)) && ($i < 10000000) ) { 
	$tries = 1;
	$numstr = sprintf("%d", $i);
	$newuser = sprintf("%-".$maxlen."s", $base);
	substr($newuser, -length($numstr), length($numstr), $numstr);
	$newuser =~ s/ //g;
        $i++;
	$debug and printf(STDERR "Checking for $newuser\n");
    }

    $i == 10000000 and return undef;
    $debug and print "final is $newuser in $tries attempts\n";
    return $newuser;
}

sub user_exists {
    my ($user, $pwdfile) = @_;

    my $retval = 0;
    my $i;
    my $buffer;

    open my $handle, "<", $pwdfile;
    while (sysread($handle, $buffer, 65536, 0)) {
	if ($buffer =~ /(^|\n)${user}:/) {
	    $retval = 1;
	    last;
	}
    }
    close $handle;
    return $retval;
}

################################################################################
sub usage {
    printf( 
"Usage: ".basename($PROGRAM_NAME)." [-h] [-d] [-mM] [-t] [basename]

Generate a new unused username suitable for assigning to a new account.

basename    If specified this will be used as the basis of the new username

            The basename will be stripped of all characters other than letters 
            and digits, and it will be truncated to no more than the maximum 
            allowed username length (8 by default, but settable with the 
            -m option).
            In the event that a user already exists by the given name a number 
            will be inserted at the end of the username and sequentially 
            increased until an available name is found.
            If basename is not specified, then \"user\" will be used as the base.
-h          Print these instructions.
-d          Enable debug messages.
-mM         Set maximum allowed username length to M (default is 8).
-t          Perform test code.  You must have a suitable passwd file in the 
            current directory in order to run this.
");
}
